#include <QtCore/QDebug>
#include <iostream>
#include "Client.h"


Client::Client(const QUrl &url, QObject *parent) : QObject(parent), m_handler(url, nullptr)
{
    connect(&m_handler, &SocketHandler::message_received, this, &Client::on_message_received);
    connect(this, &Client::message_added, &m_handler,
            &SocketHandler::on_message_added, Qt::QueuedConnection);
    connect(this, &Client::change_server_url_signal, &m_handler, &SocketHandler::on_change_server_url);
    connect(this, &Client::destroy_handler, &m_handler, &SocketHandler::on_destroy);
    connect(this, &Client::recreate_handler, &m_handler, &SocketHandler::on_recreate);
    connect(&m_handler, &SocketHandler::socket_connected, this, &Client::connected);
    connect(&m_handler, &SocketHandler::socket_disconnected, this, &Client::disconnected);

    m_handler.moveToThread(&m_handler_thread);
    m_handler_thread.start();
}


void Client::on_message_received(QString message)
{
    emit message_received(Message::from_json_string(message));
}


void Client::add_message(const Message& message)
{
    if (message.get_type() == Message::Type::LOGOUT)
    {
        m_logged_out = true;
        emit destroy_handler();
        QThread::usleep(1);
    }
    else if (m_logged_out &&
        (message.get_type() == Message::Type::LOGIN || message.get_type() == Message::Type::REGISTER))
    {
        m_logged_out = false;
        emit recreate_handler();
        QThread::usleep(1);
    }
    else
    {
        emit message_added(message.get_raw_string());
    }
}

Client::~Client()
{
    emit destroy_handler(); // this way, the handler can destroy its timer from its own thread
    m_handler_thread.quit();
    m_handler_thread.wait();
}


Client::Client(const Client &other) : QObject(other.parent()),
    m_handler(other.m_handler.url())
{
    connect(&m_handler, &SocketHandler::message_received, this, &Client::on_message_received);
    connect(this, &Client::message_added, &m_handler,
            &SocketHandler::on_message_added, Qt::QueuedConnection);
    connect(this, &Client::change_server_url_signal, &m_handler, &SocketHandler::on_change_server_url);
    connect(this, &Client::destroy_handler, &m_handler, &SocketHandler::on_destroy);
    connect(this, &Client::recreate_handler, &m_handler, &SocketHandler::on_recreate);
    connect(&m_handler, &SocketHandler::socket_connected, this, &Client::connected);
    connect(&m_handler, &SocketHandler::socket_disconnected, this, &Client::disconnected);

    m_handler.m_login_msg = other.m_handler.m_login_msg;
    m_handler.moveToThread(&m_handler_thread);
    m_handler_thread.start();
}


Client& Client::operator=(const Client& other)
{
    m_handler.m_url = other.m_handler.url();
    m_handler.m_login_msg = other.m_handler.m_login_msg;
    QMetaObject::invokeMethod(&m_handler, "try_reconnect"); // so that the command is run handler's thread

    for (const auto& msg : m_handler.unsent_messages())
    {
        on_message_received(msg);
    }

    return *this;
}


void Client::add_message_slot(QString message)
{
    add_message(Message::from_json_string(message));
}
