#include <QtCore/QCoreApplication>
#include <iostream>

#include "chatserver.h"

int main(int argc, char *argv[])
{
    QCoreApplication app(argc, argv);
    QCoreApplication::setApplicationName("chat-server");
    QCoreApplication::setApplicationVersion("0.1");
    qInstallMessageHandler([](QtMsgType type, [[maybe_unused]] const QMessageLogContext &context, const QString &msg) {
        std::cerr << QDateTime::currentDateTime().toString("yyyy-MM-ddThh:mm:ss.zzz").toStdString();
        switch (type)
        {
        case QtDebugMsg:
            std::cerr << " Debug:";
            break;
        case QtInfoMsg:
            std::cerr << " Info:";
            break;
        case QtWarningMsg:
            std::cerr << " Warning:";
            break;
        case QtCriticalMsg:
            std::cerr << " Critical:";
            break;
        case QtFatalMsg:
            std::cerr << " Fatal:";
            break;
        }
        std::cerr << " " << msg.toStdString() << "\n";
    });

    QCommandLineParser parser;
    parser.addVersionOption();
    parser.addHelpOption();
    parser.addOptions({
        {{"t", "threads"}, "Specify number of threads.", "threads"},
        {{"p", "port"}, "Specify server port.", "port"},
        {{"d", "database_path"}, "Specify path to database.", "path"}
    });

    parser.process(app);

    bool ok{};
    auto threads = parser.value("t").toUInt(&ok);
    if (!ok) {
        threads = 1;
    }
    auto port = parser.value("p").toUInt(&ok);
    if (!ok) {
        port = 1234;
    }
    auto dbpath = parser.value("d");
    if (dbpath.isEmpty()) {
        dbpath = QCoreApplication::applicationDirPath() + "/db.db";
    }
    DatabaseWrapper::setDatabasePath(dbpath);

    ChatServer server(port, threads);

    return app.exec();
}
