#include <QtCore>
#include <iostream>
#include "Tester.h"

Tester::Tester(QObject *parent) : QObject(parent),
    m_client1(m_server_url),
    m_client2(m_server_url)
{
    /*
    const QFileInfo build_dir("../../server/build/");
    if (!build_dir.exists() || !build_dir.isDir())
    {
        std::cout << "Server \'build\' directory not found, creating one..." << std::endl;
        m_server_process.start("mkdir", QStringList() << "../../server/build");
        while (!m_server_process.waitForFinished());
        m_server_process.setWorkingDirectory("../../server/build");

        std::cout << "and compiling server..." << std::endl;
        m_server_process.start("cmake", QStringList() << "../");
        while (!m_server_process.waitForFinished());
        m_server_process.start("make");
        while (!m_server_process.waitForFinished());
        std::cout << "Server compiled successfully" << std::endl;
    }
        m_server_process.setWorkingDirectory("../../server/build/");
        */
    const QFileInfo build_dir("../../build/server");
    if (!build_dir.exists() || !build_dir.isDir())
    {
        std::cout << "Server \'build\' directory not found, creating one..." << std::endl;
        m_server_process.start("mkdir", QStringList() << "../../build");
        while (!m_server_process.waitForFinished());
        m_server_process.setWorkingDirectory("../../build/");

        std::cout << "and compiling server..." << std::endl;
        m_server_process.start("cmake", QStringList() << "../");
        while (!m_server_process.waitForFinished());
        m_server_process.start("make");
        while (!m_server_process.waitForFinished());
        std::cout << "Server compiled successfully" << std::endl;
    }
    m_server_process.setWorkingDirectory("../../build/server/");

    connect(&m_client1, &Client::message_received, this, &Tester::on_client1_received);
    connect(this, &Tester::client1_send, &m_client1, &Client::add_message_slot);
    //connect(this, &Tester::client1_send, &m_client1, &Client::message_received);

    connect(&m_client2, &Client::message_received, this, &Tester::on_client2_received);
    connect(this, &Tester::client2_send, &m_client2, &Client::add_message_slot);
    //connect(this, &Tester::client2_send, &m_client2, &Client::message_received);

    m_server_process.start("./server", QStringList() << "-p " + QString::number(m_server_port));
    m_client1.moveToThread(&m_thread1);
    m_client2.moveToThread(&m_thread2);
    m_thread1.start();
    m_thread2.start();
}


void Tester::on_client1_received(Message m)
{
    m_received_messages_cl1.push_back(m);
    if (m_correct_response_cl1.empty())
    {
        std::cout << "Client1 wasn't supposed to receive message: " << m.get_raw_string().toStdString() << std::endl;
    }
    auto correct = m_correct_response_cl1.front();
    m_correct_response_cl1.pop_front();
    check_message("Client1", m, correct);

    if (m_correct_response_cl1.empty() && m_correct_response_cl2.empty())
    {
        finish();
    }
}


void Tester::on_client2_received(Message m)
{
    m_received_messages_cl2.push_back(m);
    if (m_correct_response_cl2.empty())
    {
        std::cout << "Client2 wasn't supposed to receive message: " << m.get_raw_string().toStdString() << std::endl;
    }
    auto correct = m_correct_response_cl2.front();
    m_correct_response_cl2.pop_front();
    check_message("Client2", m, correct);

    if (m_correct_response_cl1.empty() && m_correct_response_cl2.empty())
    {
        finish();
    }
}


void Tester::test_register()
{
    QString name1 = "1";
    QString password1 = "ff";
    QString name2 = "2";
    QString password2 = "ff";
    auto reg1 = Message::create_register_message(name1, password1);
    auto reg2 = Message::create_register_message(name2, password2);
    m_correct_response_cl1.push_back(Message::create_info_message("", reg1, true, "Registration successful"));
    m_correct_response_cl1.push_back(Message::create_info_message("", reg1, false, "Registration unsuccessful"));

    m_correct_response_cl2.push_back(Message::create_info_message("", reg2, true, "Registration successful"));

    std::cout << "Testing registration, this should involve 2 message from cl1 and 1 from cl2:" << std::endl;
    emit client1_send(reg1.get_raw_string());
    emit client1_send(reg1.get_raw_string());
    emit client2_send(reg2.get_raw_string());
}


void Tester::check_message(const QString &client, const Message &actual, const Message &correct)
{
    std::cout << client.toStdString() +  " received " << (actual.equals_without_time(correct) ? "correct message OK"
                                                                        : "incorrect message NOK: "
                                                                          "actual: " + actual.get_raw_string().toStdString()
                                                                          + " correct: " + correct.get_raw_string().toStdString())
              << std::endl;
}


void Tester::test_login()
{
    auto incorrect_password = Message::create_login_message(cl1_name, cl1_pass + "@");
    m_correct_response_cl1.push_back(Message::create_info_message("", incorrect_password, false, "Login unsuccessful"));
    auto name_not_in_database = Message::create_login_message(cl1_name + "XXX" + cl2_name, cl1_pass);
    m_correct_response_cl2.push_back(Message::create_info_message("", name_not_in_database, false, "Login unsuccessful"));

    auto log1 = Message::create_login_message(cl1_name, cl1_pass);
    m_correct_response_cl1.push_back(Message::create_info_message("", log1, true, "Login successful"));
    m_correct_response_cl1.push_back(Message::create_contact_list_response(QJsonArray()));
    auto log2 = Message::create_login_message(cl2_name, cl2_pass); // both correct
    m_correct_response_cl2.push_back(Message::create_info_message("", log2, true, "Login successful"));
    m_correct_response_cl2.push_back(Message::create_contact_list_response(QJsonArray()));
    std::cout << "Client login, this should involve 3 messages from cl1 and 3 from cl2:" << std::endl;


    emit client1_send(incorrect_password.get_raw_string());
    emit client2_send(name_not_in_database.get_raw_string());
    emit client1_send(log1.get_raw_string());
    emit client2_send(log2.get_raw_string());
}


void Tester::test_friend_request()
{
    auto req = Message::create_friend_request(cl1_name, cl2_name);
    m_correct_response_cl1.push_back(Message::create_contact_list_response(QJsonArray() << cl2_name));
    m_correct_response_cl2.push_back(Message::create_contact_list_response(QJsonArray() << cl1_name));

    emit client1_send(req.get_raw_string());
}


void Tester::test_message_sending()
{
    auto msg1 = Message::create_text_message(cl1_name, cl2_name, "Hi");
    m_correct_response_cl2.push_back(msg1);
    emit client1_send(msg1.get_raw_string());


    auto msg2 = Message::create_text_message(cl2_name, cl1_name, "Hi");
    m_correct_response_cl1.push_back(msg2);
    emit client2_send(msg2.get_raw_string());
}

void Tester::register_clients()
{
    auto reg1 = Message::create_register_message(cl1_name, cl1_pass);
    auto reg2 = Message::create_register_message(cl2_name, cl2_pass);
    m_correct_response_cl1.push_back(Message::create_info_message("", reg1, true, "Registration successful"));
    m_correct_response_cl2.push_back(Message::create_info_message("", reg2, true, "Registration successful"));
    std::cout << "Client registration, this should involve 1 message from cl1 and 1 from cl2:" << std::endl;


    emit client1_send(reg1.get_raw_string());
    emit client2_send(reg2.get_raw_string());
}

void Tester::login_clients()
{
    auto reg1 = Message::create_login_message(cl1_name, cl1_pass);
    auto reg2 = Message::create_login_message(cl2_name, cl2_pass);
    m_correct_response_cl1.push_back(Message::create_info_message("", reg1, true, "Login successful"));
    m_correct_response_cl1.push_back(Message::create_contact_list_response(QJsonArray()));
    m_correct_response_cl2.push_back(Message::create_info_message("", reg2, true, "Login successful"));
    m_correct_response_cl2.push_back(Message::create_contact_list_response(QJsonArray()));

    emit client1_send(reg1.get_raw_string());
    emit client2_send(reg2.get_raw_string());
}

void Tester::test_all()
{
    test_register();

    register_clients();
    test_login();

    QThread::sleep(6); // we dont want to update friends before the first login
    test_friend_request();
    test_message_sending();
}

void Tester::finish()
{
    m_thread1.quit();
    m_thread2.quit();
    m_thread1.wait();
    m_thread2.wait();
    std::cout << "Tests finished" << std::endl;
    QProcess p;
    p.setWorkingDirectory("../../build/server/");
    p.start("rm", QStringList() << "db.db");
    p.waitForFinished();
    QMetaObject::invokeMethod(&m_server_process, "kill"); // so that the command is run in respective thread
}
